# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
# @@protoc_insertion_point(imports)



DESCRIPTOR = descriptor.FileDescriptor(
  name='distributions/io/schema.proto',
  package='protobuf.distributions',
  serialized_pb='\n\x1d\x64istributions/io/schema.proto\x12\x16protobuf.distributions\"\xdd\x01\n\nClustering\x12@\n\npitman_yor\x18\x01 \x01(\x0b\x32,.protobuf.distributions.Clustering.PitmanYor\x12\x42\n\x0blow_entropy\x18\x02 \x01(\x0b\x32-.protobuf.distributions.Clustering.LowEntropy\x1a%\n\tPitmanYor\x12\r\n\x05\x61lpha\x18\x01 \x02(\x02\x12\t\n\x01\x64\x18\x02 \x02(\x02\x1a\"\n\nLowEntropy\x12\x14\n\x0c\x64\x61taset_size\x18\x01 \x02(\x04\"\x87\x01\n\rBetaBernoulli\x1a%\n\x06Shared\x12\r\n\x05\x61lpha\x18\x01 \x02(\x02\x12\x0c\n\x04\x62\x65ta\x18\x02 \x02(\x02\x1a(\n\tGridPrior\x12\r\n\x05\x61lpha\x18\x01 \x03(\x02\x12\x0c\n\x04\x62\x65ta\x18\x02 \x03(\x02\x1a%\n\x05Group\x12\r\n\x05heads\x18\x01 \x02(\x04\x12\r\n\x05tails\x18\x02 \x02(\x04\"b\n\x11\x44irichletDiscrete\x1a\x18\n\x06Shared\x12\x0e\n\x06\x61lphas\x18\x01 \x03(\x02\x1a\x1a\n\tGridPrior\x12\r\n\x05\x61lpha\x18\x01 \x03(\x02\x1a\x17\n\x05Group\x12\x0e\n\x06\x63ounts\x18\x01 \x03(\x04\"\xc3\x01\n\x18\x44irichletProcessDiscrete\x1aU\n\x06Shared\x12\r\n\x05gamma\x18\x01 \x02(\x02\x12\r\n\x05\x61lpha\x18\x02 \x02(\x02\x12\x0e\n\x06values\x18\x03 \x03(\r\x12\r\n\x05\x62\x65tas\x18\x04 \x03(\x02\x12\x0e\n\x06\x63ounts\x18\x05 \x03(\x04\x1a)\n\tGridPrior\x12\r\n\x05gamma\x18\x01 \x03(\x02\x12\r\n\x05\x61lpha\x18\x02 \x03(\x02\x1a%\n\x05Group\x12\x0c\n\x04keys\x18\x01 \x03(\r\x12\x0e\n\x06values\x18\x02 \x03(\x04\"\xc1\x01\n\x18PitmanYorProcessDiscrete\x1a\x32\n\x06Shared\x12\r\n\x05\x61lpha\x18\x01 \x02(\x02\x12\t\n\x01\x64\x18\x02 \x03(\x02\x12\x0e\n\x06\x63ounts\x18\x03 \x03(\x04\x1aJ\n\tGridPrior\x12=\n\x07\x61lpha_d\x18\x01 \x03(\x0b\x32,.protobuf.distributions.Clustering.PitmanYor\x1a%\n\x05Group\x12\x0c\n\x04keys\x18\x01 \x03(\r\x12\x0e\n\x06values\x18\x02 \x03(\x04\"\x9e\x01\n\x0cGammaPoisson\x1a)\n\x06Shared\x12\r\n\x05\x61lpha\x18\x01 \x02(\x02\x12\x10\n\x08inv_beta\x18\x02 \x02(\x02\x1a,\n\tGridPrior\x12\r\n\x05\x61lpha\x18\x01 \x03(\x02\x12\x10\n\x08inv_beta\x18\x02 \x03(\x02\x1a\x35\n\x05Group\x12\r\n\x05\x63ount\x18\x01 \x02(\x04\x12\x0b\n\x03sum\x18\x02 \x02(\x04\x12\x10\n\x08log_prod\x18\x03 \x02(\x02\"\xa2\x01\n\x14\x42\x65taNegativeBinomial\x1a\x30\n\x06Shared\x12\r\n\x05\x61lpha\x18\x01 \x02(\x02\x12\x0c\n\x04\x62\x65ta\x18\x02 \x02(\x02\x12\t\n\x01r\x18\x03 \x02(\x04\x1a\x33\n\tGridPrior\x12\r\n\x05\x61lpha\x18\x01 \x03(\x02\x12\x0c\n\x04\x62\x65ta\x18\x02 \x03(\x02\x12\t\n\x01r\x18\x03 \x03(\x04\x1a#\n\x05Group\x12\r\n\x05\x63ount\x18\x01 \x02(\x04\x12\x0b\n\x03sum\x18\x02 \x02(\x04\"\xdf\x01\n\x12NormalInverseChiSq\x1a@\n\x06Shared\x12\n\n\x02mu\x18\x01 \x02(\x02\x12\r\n\x05kappa\x18\x02 \x02(\x02\x12\x0f\n\x07sigmasq\x18\x03 \x02(\x02\x12\n\n\x02nu\x18\x04 \x02(\x02\x1a\x43\n\tGridPrior\x12\n\n\x02mu\x18\x01 \x03(\x02\x12\r\n\x05kappa\x18\x02 \x03(\x02\x12\x0f\n\x07sigmasq\x18\x03 \x03(\x02\x12\n\n\x02nu\x18\x04 \x03(\x02\x1a\x42\n\x05Group\x12\r\n\x05\x63ount\x18\x01 \x02(\x04\x12\x0c\n\x04mean\x18\x02 \x02(\x02\x12\x1c\n\x14\x63ount_times_variance\x18\x03 \x02(\x02')




_CLUSTERING_PITMANYOR = descriptor.Descriptor(
  name='PitmanYor',
  full_name='protobuf.distributions.Clustering.PitmanYor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.Clustering.PitmanYor.alpha', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='d', full_name='protobuf.distributions.Clustering.PitmanYor.d', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=206,
  serialized_end=243,
)

_CLUSTERING_LOWENTROPY = descriptor.Descriptor(
  name='LowEntropy',
  full_name='protobuf.distributions.Clustering.LowEntropy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='dataset_size', full_name='protobuf.distributions.Clustering.LowEntropy.dataset_size', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=245,
  serialized_end=279,
)

_CLUSTERING = descriptor.Descriptor(
  name='Clustering',
  full_name='protobuf.distributions.Clustering',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='pitman_yor', full_name='protobuf.distributions.Clustering.pitman_yor', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='low_entropy', full_name='protobuf.distributions.Clustering.low_entropy', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_CLUSTERING_PITMANYOR, _CLUSTERING_LOWENTROPY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=58,
  serialized_end=279,
)


_BETABERNOULLI_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.BetaBernoulli.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.BetaBernoulli.Shared.alpha', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='beta', full_name='protobuf.distributions.BetaBernoulli.Shared.beta', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=299,
  serialized_end=336,
)

_BETABERNOULLI_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.BetaBernoulli.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.BetaBernoulli.GridPrior.alpha', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='beta', full_name='protobuf.distributions.BetaBernoulli.GridPrior.beta', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=338,
  serialized_end=378,
)

_BETABERNOULLI_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.BetaBernoulli.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='heads', full_name='protobuf.distributions.BetaBernoulli.Group.heads', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='tails', full_name='protobuf.distributions.BetaBernoulli.Group.tails', index=1,
      number=2, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=380,
  serialized_end=417,
)

_BETABERNOULLI = descriptor.Descriptor(
  name='BetaBernoulli',
  full_name='protobuf.distributions.BetaBernoulli',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_BETABERNOULLI_SHARED, _BETABERNOULLI_GRIDPRIOR, _BETABERNOULLI_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=282,
  serialized_end=417,
)


_DIRICHLETDISCRETE_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.DirichletDiscrete.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alphas', full_name='protobuf.distributions.DirichletDiscrete.Shared.alphas', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=440,
  serialized_end=464,
)

_DIRICHLETDISCRETE_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.DirichletDiscrete.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.DirichletDiscrete.GridPrior.alpha', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=338,
  serialized_end=364,
)

_DIRICHLETDISCRETE_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.DirichletDiscrete.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='counts', full_name='protobuf.distributions.DirichletDiscrete.Group.counts', index=0,
      number=1, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=494,
  serialized_end=517,
)

_DIRICHLETDISCRETE = descriptor.Descriptor(
  name='DirichletDiscrete',
  full_name='protobuf.distributions.DirichletDiscrete',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_DIRICHLETDISCRETE_SHARED, _DIRICHLETDISCRETE_GRIDPRIOR, _DIRICHLETDISCRETE_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=419,
  serialized_end=517,
)


_DIRICHLETPROCESSDISCRETE_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.DirichletProcessDiscrete.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='gamma', full_name='protobuf.distributions.DirichletProcessDiscrete.Shared.gamma', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.DirichletProcessDiscrete.Shared.alpha', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='values', full_name='protobuf.distributions.DirichletProcessDiscrete.Shared.values', index=2,
      number=3, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='betas', full_name='protobuf.distributions.DirichletProcessDiscrete.Shared.betas', index=3,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='counts', full_name='protobuf.distributions.DirichletProcessDiscrete.Shared.counts', index=4,
      number=5, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=548,
  serialized_end=633,
)

_DIRICHLETPROCESSDISCRETE_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.DirichletProcessDiscrete.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='gamma', full_name='protobuf.distributions.DirichletProcessDiscrete.GridPrior.gamma', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.DirichletProcessDiscrete.GridPrior.alpha', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=635,
  serialized_end=676,
)

_DIRICHLETPROCESSDISCRETE_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.DirichletProcessDiscrete.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='keys', full_name='protobuf.distributions.DirichletProcessDiscrete.Group.keys', index=0,
      number=1, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='values', full_name='protobuf.distributions.DirichletProcessDiscrete.Group.values', index=1,
      number=2, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=678,
  serialized_end=715,
)

_DIRICHLETPROCESSDISCRETE = descriptor.Descriptor(
  name='DirichletProcessDiscrete',
  full_name='protobuf.distributions.DirichletProcessDiscrete',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_DIRICHLETPROCESSDISCRETE_SHARED, _DIRICHLETPROCESSDISCRETE_GRIDPRIOR, _DIRICHLETPROCESSDISCRETE_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=520,
  serialized_end=715,
)


_PITMANYORPROCESSDISCRETE_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.PitmanYorProcessDiscrete.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.PitmanYorProcessDiscrete.Shared.alpha', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='d', full_name='protobuf.distributions.PitmanYorProcessDiscrete.Shared.d', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='counts', full_name='protobuf.distributions.PitmanYorProcessDiscrete.Shared.counts', index=2,
      number=3, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=746,
  serialized_end=796,
)

_PITMANYORPROCESSDISCRETE_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.PitmanYorProcessDiscrete.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha_d', full_name='protobuf.distributions.PitmanYorProcessDiscrete.GridPrior.alpha_d', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=798,
  serialized_end=872,
)

_PITMANYORPROCESSDISCRETE_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.PitmanYorProcessDiscrete.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='keys', full_name='protobuf.distributions.PitmanYorProcessDiscrete.Group.keys', index=0,
      number=1, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='values', full_name='protobuf.distributions.PitmanYorProcessDiscrete.Group.values', index=1,
      number=2, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=678,
  serialized_end=715,
)

_PITMANYORPROCESSDISCRETE = descriptor.Descriptor(
  name='PitmanYorProcessDiscrete',
  full_name='protobuf.distributions.PitmanYorProcessDiscrete',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_PITMANYORPROCESSDISCRETE_SHARED, _PITMANYORPROCESSDISCRETE_GRIDPRIOR, _PITMANYORPROCESSDISCRETE_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=718,
  serialized_end=911,
)


_GAMMAPOISSON_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.GammaPoisson.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.GammaPoisson.Shared.alpha', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='inv_beta', full_name='protobuf.distributions.GammaPoisson.Shared.inv_beta', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=930,
  serialized_end=971,
)

_GAMMAPOISSON_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.GammaPoisson.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.GammaPoisson.GridPrior.alpha', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='inv_beta', full_name='protobuf.distributions.GammaPoisson.GridPrior.inv_beta', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=973,
  serialized_end=1017,
)

_GAMMAPOISSON_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.GammaPoisson.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='count', full_name='protobuf.distributions.GammaPoisson.Group.count', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='sum', full_name='protobuf.distributions.GammaPoisson.Group.sum', index=1,
      number=2, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='log_prod', full_name='protobuf.distributions.GammaPoisson.Group.log_prod', index=2,
      number=3, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1019,
  serialized_end=1072,
)

_GAMMAPOISSON = descriptor.Descriptor(
  name='GammaPoisson',
  full_name='protobuf.distributions.GammaPoisson',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_GAMMAPOISSON_SHARED, _GAMMAPOISSON_GRIDPRIOR, _GAMMAPOISSON_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=914,
  serialized_end=1072,
)


_BETANEGATIVEBINOMIAL_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.BetaNegativeBinomial.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.BetaNegativeBinomial.Shared.alpha', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='beta', full_name='protobuf.distributions.BetaNegativeBinomial.Shared.beta', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='r', full_name='protobuf.distributions.BetaNegativeBinomial.Shared.r', index=2,
      number=3, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1099,
  serialized_end=1147,
)

_BETANEGATIVEBINOMIAL_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.BetaNegativeBinomial.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.BetaNegativeBinomial.GridPrior.alpha', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='beta', full_name='protobuf.distributions.BetaNegativeBinomial.GridPrior.beta', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='r', full_name='protobuf.distributions.BetaNegativeBinomial.GridPrior.r', index=2,
      number=3, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1149,
  serialized_end=1200,
)

_BETANEGATIVEBINOMIAL_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.BetaNegativeBinomial.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='count', full_name='protobuf.distributions.BetaNegativeBinomial.Group.count', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='sum', full_name='protobuf.distributions.BetaNegativeBinomial.Group.sum', index=1,
      number=2, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1019,
  serialized_end=1054,
)

_BETANEGATIVEBINOMIAL = descriptor.Descriptor(
  name='BetaNegativeBinomial',
  full_name='protobuf.distributions.BetaNegativeBinomial',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_BETANEGATIVEBINOMIAL_SHARED, _BETANEGATIVEBINOMIAL_GRIDPRIOR, _BETANEGATIVEBINOMIAL_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1075,
  serialized_end=1237,
)


_NORMALINVERSECHISQ_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.NormalInverseChiSq.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='mu', full_name='protobuf.distributions.NormalInverseChiSq.Shared.mu', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='kappa', full_name='protobuf.distributions.NormalInverseChiSq.Shared.kappa', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='sigmasq', full_name='protobuf.distributions.NormalInverseChiSq.Shared.sigmasq', index=2,
      number=3, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='nu', full_name='protobuf.distributions.NormalInverseChiSq.Shared.nu', index=3,
      number=4, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1262,
  serialized_end=1326,
)

_NORMALINVERSECHISQ_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.NormalInverseChiSq.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='mu', full_name='protobuf.distributions.NormalInverseChiSq.GridPrior.mu', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='kappa', full_name='protobuf.distributions.NormalInverseChiSq.GridPrior.kappa', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='sigmasq', full_name='protobuf.distributions.NormalInverseChiSq.GridPrior.sigmasq', index=2,
      number=3, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='nu', full_name='protobuf.distributions.NormalInverseChiSq.GridPrior.nu', index=3,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1328,
  serialized_end=1395,
)

_NORMALINVERSECHISQ_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.NormalInverseChiSq.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='count', full_name='protobuf.distributions.NormalInverseChiSq.Group.count', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='mean', full_name='protobuf.distributions.NormalInverseChiSq.Group.mean', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='count_times_variance', full_name='protobuf.distributions.NormalInverseChiSq.Group.count_times_variance', index=2,
      number=3, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1397,
  serialized_end=1463,
)

_NORMALINVERSECHISQ = descriptor.Descriptor(
  name='NormalInverseChiSq',
  full_name='protobuf.distributions.NormalInverseChiSq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_NORMALINVERSECHISQ_SHARED, _NORMALINVERSECHISQ_GRIDPRIOR, _NORMALINVERSECHISQ_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1240,
  serialized_end=1463,
)

_CLUSTERING_PITMANYOR.containing_type = _CLUSTERING;
_CLUSTERING_LOWENTROPY.containing_type = _CLUSTERING;
_CLUSTERING.fields_by_name['pitman_yor'].message_type = _CLUSTERING_PITMANYOR
_CLUSTERING.fields_by_name['low_entropy'].message_type = _CLUSTERING_LOWENTROPY
_BETABERNOULLI_SHARED.containing_type = _BETABERNOULLI;
_BETABERNOULLI_GRIDPRIOR.containing_type = _BETABERNOULLI;
_BETABERNOULLI_GROUP.containing_type = _BETABERNOULLI;
_DIRICHLETDISCRETE_SHARED.containing_type = _DIRICHLETDISCRETE;
_DIRICHLETDISCRETE_GRIDPRIOR.containing_type = _DIRICHLETDISCRETE;
_DIRICHLETDISCRETE_GROUP.containing_type = _DIRICHLETDISCRETE;
_DIRICHLETPROCESSDISCRETE_SHARED.containing_type = _DIRICHLETPROCESSDISCRETE;
_DIRICHLETPROCESSDISCRETE_GRIDPRIOR.containing_type = _DIRICHLETPROCESSDISCRETE;
_DIRICHLETPROCESSDISCRETE_GROUP.containing_type = _DIRICHLETPROCESSDISCRETE;
_PITMANYORPROCESSDISCRETE_SHARED.containing_type = _PITMANYORPROCESSDISCRETE;
_PITMANYORPROCESSDISCRETE_GRIDPRIOR.fields_by_name['alpha_d'].message_type = _CLUSTERING_PITMANYOR
_PITMANYORPROCESSDISCRETE_GRIDPRIOR.containing_type = _PITMANYORPROCESSDISCRETE;
_PITMANYORPROCESSDISCRETE_GROUP.containing_type = _PITMANYORPROCESSDISCRETE;
_GAMMAPOISSON_SHARED.containing_type = _GAMMAPOISSON;
_GAMMAPOISSON_GRIDPRIOR.containing_type = _GAMMAPOISSON;
_GAMMAPOISSON_GROUP.containing_type = _GAMMAPOISSON;
_BETANEGATIVEBINOMIAL_SHARED.containing_type = _BETANEGATIVEBINOMIAL;
_BETANEGATIVEBINOMIAL_GRIDPRIOR.containing_type = _BETANEGATIVEBINOMIAL;
_BETANEGATIVEBINOMIAL_GROUP.containing_type = _BETANEGATIVEBINOMIAL;
_NORMALINVERSECHISQ_SHARED.containing_type = _NORMALINVERSECHISQ;
_NORMALINVERSECHISQ_GRIDPRIOR.containing_type = _NORMALINVERSECHISQ;
_NORMALINVERSECHISQ_GROUP.containing_type = _NORMALINVERSECHISQ;
DESCRIPTOR.message_types_by_name['Clustering'] = _CLUSTERING
DESCRIPTOR.message_types_by_name['BetaBernoulli'] = _BETABERNOULLI
DESCRIPTOR.message_types_by_name['DirichletDiscrete'] = _DIRICHLETDISCRETE
DESCRIPTOR.message_types_by_name['DirichletProcessDiscrete'] = _DIRICHLETPROCESSDISCRETE
DESCRIPTOR.message_types_by_name['PitmanYorProcessDiscrete'] = _PITMANYORPROCESSDISCRETE
DESCRIPTOR.message_types_by_name['GammaPoisson'] = _GAMMAPOISSON
DESCRIPTOR.message_types_by_name['BetaNegativeBinomial'] = _BETANEGATIVEBINOMIAL
DESCRIPTOR.message_types_by_name['NormalInverseChiSq'] = _NORMALINVERSECHISQ

class Clustering(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class PitmanYor(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CLUSTERING_PITMANYOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.Clustering.PitmanYor)
  
  class LowEntropy(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CLUSTERING_LOWENTROPY
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.Clustering.LowEntropy)
  DESCRIPTOR = _CLUSTERING
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.Clustering)

class BetaBernoulli(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETABERNOULLI_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaBernoulli.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETABERNOULLI_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaBernoulli.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETABERNOULLI_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaBernoulli.Group)
  DESCRIPTOR = _BETABERNOULLI
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaBernoulli)

class DirichletDiscrete(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETDISCRETE_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletDiscrete.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETDISCRETE_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletDiscrete.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETDISCRETE_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletDiscrete.Group)
  DESCRIPTOR = _DIRICHLETDISCRETE
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletDiscrete)

class DirichletProcessDiscrete(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETPROCESSDISCRETE_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletProcessDiscrete.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETPROCESSDISCRETE_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletProcessDiscrete.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETPROCESSDISCRETE_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletProcessDiscrete.Group)
  DESCRIPTOR = _DIRICHLETPROCESSDISCRETE
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletProcessDiscrete)

class PitmanYorProcessDiscrete(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _PITMANYORPROCESSDISCRETE_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.PitmanYorProcessDiscrete.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _PITMANYORPROCESSDISCRETE_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.PitmanYorProcessDiscrete.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _PITMANYORPROCESSDISCRETE_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.PitmanYorProcessDiscrete.Group)
  DESCRIPTOR = _PITMANYORPROCESSDISCRETE
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.PitmanYorProcessDiscrete)

class GammaPoisson(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _GAMMAPOISSON_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.GammaPoisson.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _GAMMAPOISSON_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.GammaPoisson.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _GAMMAPOISSON_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.GammaPoisson.Group)
  DESCRIPTOR = _GAMMAPOISSON
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.GammaPoisson)

class BetaNegativeBinomial(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETANEGATIVEBINOMIAL_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaNegativeBinomial.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETANEGATIVEBINOMIAL_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaNegativeBinomial.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETANEGATIVEBINOMIAL_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaNegativeBinomial.Group)
  DESCRIPTOR = _BETANEGATIVEBINOMIAL
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaNegativeBinomial)

class NormalInverseChiSq(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _NORMALINVERSECHISQ_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.NormalInverseChiSq.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _NORMALINVERSECHISQ_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.NormalInverseChiSq.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _NORMALINVERSECHISQ_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.NormalInverseChiSq.Group)
  DESCRIPTOR = _NORMALINVERSECHISQ
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.NormalInverseChiSq)

# @@protoc_insertion_point(module_scope)
